class TestVersion:
    def test_version_is_consistent(self):
        from .. import VERSION

        with open('README.md') as f:
            readme = f.read().split('\n')
            version_list = readme[3].split(':')[1].strip().split('.')
            version_list = [int(i) for i in version_list]
            readme_version = tuple(version_list)

        assert VERSION == readme_version



import socket
from threading import Lock
from collections import defaultdict, deque
from queue import Queue
import struct
import logging

from .concurrency import synchronized
from .message import Message
from .exceptions import UnexpectedFrame, Timeout, METHOD_NAME_MAP
from .serialization import AMQPReader
from . import spec
from .spec import FrameType, Frame, Method, method_t

log = logging.getLogger('amqpy')

__all__ = ['MethodReader']

# these methods are followed by content headers and bodies
_CONTENT_METHODS = [
    spec.Basic.Return,
    spec.Basic.Deliver,
    spec.Basic.GetOk,
]


class PartialMessage:
    """Helper class to build up a multi-frame `Method` with a complete `Message`
    """

    def __init__(self, method_type, args, channel):
        """
        :param method_type: method type
        :param args: method args
        :param channel: associated channel
        :type method_type: method_t
        :type args: AMQPReader, AMQPWriter
        :type channel: int
        """
        self.method_type = method_type
        self.args = args
        self.channel = channel
        self.msg = Message()
        self.body_parts = bytearray()
        self.expected_body_size = None

    def add_header(self, payload):
        """Add header to partial message

        :param payload: frame payload of a `FrameType.HEADER` frame
        :type payload: bytes
        """
        class_id, weight, self.expected_body_size = struct.unpack('>HHQ', payload[:12])
        self.msg.load_properties(payload[12:])

    def add_payload(self, payload):
        """Add content to partial message

        :param payload: frame payload of a `FrameType.BODY` frame
        :type payload: bytes
        """
        self.body_parts.extend(payload)
        if self.complete:
            # TODO: should we bother converting this to an immutable bytes object just to make tests pass?
            self.msg.body = bytes(self.body_parts)

    @property
    def complete(self):
        """Check if message is complete

        :return: message complete?
        :rtype: bool
        """
        return self.expected_body_size == 0 or len(self.body_parts) == self.expected_body_size


class MethodReader:
    """Read frames from the server and construct complete methods

    There should be one `MethodReader` instance per connection.

    In the case of a framing error, an :exc:`AMQPConnectionError` is placed in the queue.
    In the case of unexpected frames, an :exc:`ChannelError` is placed in the queue.
    """

    def __init__(self, source):
        """
        :param source: connection source transport
        :type source: amqpy.transport.TCPTransport or amqpy.transport.SSLTransport
        """
        self.source = source
        self.sock = source.sock

        # deque[Method or Exception]
        self.method_queue = deque()
        # dict[channel_id int: PartialMessage]
        self.partial_messages = {}
        # dict[channel_id int: frame_type int]
        self.expected_types = defaultdict(lambda: FrameType.METHOD)  # next expected frame type for each channel

        self.heartbeats = 0  # total number of heartbeats received
        self.frames_recv = 0  # total number of frames received

        self._method_read_lock = Lock()

    def _next_method(self):
        """Read the next method from the source and process it

        Once one complete method has been assembled, it is placed in the internal queue. This method will block until a
        complete `Method` has been constructed, which may consist of one or more frames.
        """
        while not self.method_queue:
            # keep reading frames until we have at least one complete method in the queue
            try:
                frame = self.source.read_frame()
            except Exception as exc:
                # connection was closed? framing error?
                self.method_queue.append(exc)
                break

            self.frames_recv += 1

            if frame.frame_type not in (self.expected_types[frame.channel], 8):
                msg = 'Received frame type {} while expecting type: {}' \
                    .format(frame.frame_type, self.expected_types[frame.channel])
                self.method_queue.append(UnexpectedFrame(msg, channel_id=frame.channel))
            elif frame.frame_type == FrameType.METHOD:
                self._process_method_frame(frame)
            elif frame.frame_type == FrameType.HEADER:
                self._process_content_header(frame)
            elif frame.frame_type == FrameType.BODY:
                self._process_content_body(frame)
            elif frame.frame_type == FrameType.HEARTBEAT:
                self._process_heartbeat()

    def _process_heartbeat(self):
        self.heartbeats += 1

    def _process_method_frame(self, frame):
        """Process method frame
        """
        channel = frame.channel
        # noinspection PyTypeChecker
        method_type = method_t(*struct.unpack('>HH', frame.payload[:4]))
        args = AMQPReader(frame.payload[4:])

        if method_type in _CONTENT_METHODS:
            # save what we've got so far and wait for the content-header
            self.partial_messages[channel] = PartialMessage(method_type, args, channel)
            self.expected_types[channel] = spec.FrameType.HEADER
        else:
            # this is a complete method
            method = Method(method_type, args, None, channel)
            self.method_queue.append(method)

    def _process_content_header(self, frame):
        """Process Content Header frames
        """
        partial = self.partial_messages[frame.channel]
        partial.add_header(frame.payload)

        if partial.complete:
            # a bodyless message, we're done
            method = Method(partial.method_type, partial.args, partial.msg, frame.channel)
            self.method_queue.append(method)
            self.partial_messages.pop(frame.channel, None)
            del self.expected_types[frame.channel]  # reset expected frame type for this channel
        else:
            # next expected frame type is FrameType.BODY
            self.expected_types[frame.channel] = spec.FrameType.BODY

    def _process_content_body(self, frame):
        """Process Content Body frames
        """
        partial = self.partial_messages[frame.channel]
        partial.add_payload(frame.payload)

        if partial.complete:
            # message is complete, append it to the queue
            method = Method(partial.method_type, partial.args, partial.msg, frame.channel)
            self.method_queue.append(method)
            self.partial_messages.pop(frame.channel, None)
            del self.expected_types[frame.channel]  # reset expected frame type for this channel

    @synchronized('_method_read_lock')
    def _read_method(self):
        """Read a method from the peer

        :return: method
        :rtype: amqpy.spec.Method
        """
        # fully read and process next method
        self._next_method()
        method = self.method_queue.popleft()

        # `method` may sometimes be an `Exception`, raise it here
        if isinstance(method, Exception):
            raise method

        log.debug('{:7} {} {}'.format('Read:', method.method_type, METHOD_NAME_MAP[method.method_type]))
        return method

    def read_method(self, timeout=None):
        """Read method

        :param timeout: timeout
        :type timeout: float
        :return: method
        :rtype: amqpy.spec.Method
        :raise amqpy.exceptions.Timeout: if the operation times out
        """
        if timeout is None:
            return self._read_method()

        orig_timeout = self.sock.gettimeout()
        if orig_timeout != timeout:
            self.sock.settimeout(timeout)

        try:
            return self._read_method()
        except socket.timeout:
            raise Timeout()
        finally:
            if orig_timeout != timeout:
                self.sock.settimeout(orig_timeout)


class MethodWriter:
    """Write methods to the server by breaking them up and constructing multiple frames

    There should be one `MethodWriter` instance per connection, and all channels share that instance. This class is
    thread-safe. Any thread may call :meth:`write_method()` as long as no more than one thread is writing to any
    given `channel_id` at a time.
    """

    def __init__(self, dest, frame_max):
        """
        :param dest: destination transport
        :param frame_max: maximum frame payload size in bytes
        :type dest: amqpy.transport.AbstractTransport
        :type frame_max: int
        """
        self.dest = dest
        self.frame_max = frame_max
        self.methods_sent = 0  # total number of methods sent

    def write_method(self, channel_id, method):
        """Write method to connection, destined for the specified `channel_id`

        This implementation uses a queue internally to prepare all frames before writing in order to detect issues.

        This method is thread safe only if the `channel_id` parameter is unique across concurrent invocations. The AMQP
        protocol allows interleaving frames destined for different channels, but not within the same channel. This means
        no more than one thread may safely operate on any given channel.

        :param channel_id: channel
        :param method: method to write
        :type channel_id: int
        :type method: Method
        """
        log.debug('{:7} {} {}'.format('Write:', method.method_type, METHOD_NAME_MAP[method.method_type]))
        frames = Queue()

        # construct a method frame
        frame = Frame(FrameType.METHOD, channel_id, method.pack_method())
        frames.put(frame)

        if method.content:
            # construct a header frame
            frame = Frame(FrameType.HEADER, channel_id, method.pack_header())
            frames.put(frame)

            # construct one or more body frames, which contain the body of the `Message`
            chunk_size = self.frame_max - 8
            for payload in method.pack_body(chunk_size):
                frame = Frame(FrameType.BODY, channel_id, payload)
                frames.put(frame)

        while not frames.empty():
            self.dest.write_frame(frames.get())

        self.methods_sent += 1



"""Convert between bytestreams and higher-level AMQP types
"""
import io
from datetime import datetime
from decimal import Decimal
from struct import pack, unpack
from time import mktime

from .exceptions import FrameSyntaxError


def byte(n):
    return bytes([n])


class AMQPReader:
    """Read higher-level AMQP types from a bytestream
    """

    def __init__(self, source):
        """
        :param source: source bytes or file-like object
        :type source: io.BytesIO or bytes or bytearray
        """
        if isinstance(source, bytes) or isinstance(source, bytearray):
            self.input = io.BytesIO(source)
        elif isinstance(source, io.BytesIO):
            self.input = source
        else:
            raise TypeError('AMQPReader needs an `io.BytesIO` or `bytes` or `bytearray`')

        self.bit_count = self.bits = 0

    def close(self):
        self.input.close()

    def getvalue(self):
        return self.input.getvalue()

    def read(self, n=-1):
        """Read n bytes
        """
        self.bit_count = self.bits = 0
        return self.input.read(n)

    def read_bit(self):
        """Read a single boolean value
        """
        if not self.bit_count:
            self.bits = ord(self.input.read(1))
            self.bit_count = 8
        result = (self.bits & 1) == 1
        self.bits >>= 1
        self.bit_count -= 1
        return result

    def read_octet(self):
        """Read one byte, return as an integer
        """
        self.bit_count = self.bits = 0
        return unpack('B', self.input.read(1))[0]

    def read_short(self):
        """Read an unsigned 16-bit integer
        """
        self.bit_count = self.bits = 0
        return unpack('>H', self.input.read(2))[0]

    def read_long(self):
        """Read an unsigned 32-bit integer
        """
        self.bit_count = self.bits = 0
        return unpack('>I', self.input.read(4))[0]

    def read_longlong(self):
        """Read an unsigned 64-bit integer
        """
        self.bit_count = self.bits = 0
        return unpack('>Q', self.input.read(8))[0]

    def read_float(self):
        """Read float value."""
        self.bit_count = self.bits = 0
        return unpack('>d', self.input.read(8))[0]

    def read_shortstr(self):
        """Read a short string that's stored in up to 255 bytes

        The encoding isn't specified in the AMQP spec, so assume it's utf-8
        """
        self.bit_count = self.bits = 0
        slen = unpack('B', self.input.read(1))[0]
        return self.input.read(slen).decode('utf-8')

    def read_longstr(self):
        """Read a string that's up to 2**32 bytes

        The encoding isn't specified in the AMQP spec, so assume it's utf-8
        """
        self.bit_count = self.bits = 0
        slen = unpack('>I', self.input.read(4))[0]
        return self.input.read(slen).decode('utf-8')

    def read_table(self):
        """Read an AMQP table, and return as a Python dictionary
        """
        self.bit_count = self.bits = 0
        tlen = unpack('>I', self.input.read(4))[0]
        table_data = AMQPReader(self.input.read(tlen))
        result = {}
        while table_data.input.tell() < tlen:
            name = table_data.read_shortstr()
            val = table_data.read_item()
            result[name] = val
        return result

    def read_item(self):
        ftype = ord(self.input.read(1))

        # 'S': long string
        if ftype == 83:
            val = self.read_longstr()
        # 's': short string
        elif ftype == 115:
            val = self.read_shortstr()
        # 'b': short-short int
        elif ftype == 98:
            val, = unpack('>B', self.input.read(1))
        # 'B': short-short unsigned int
        elif ftype == 66:
            val, = unpack('>b', self.input.read(1))
        # 'U': short int
        elif ftype == 85:
            val, = unpack('>h', self.input.read(2))
        # 'u': short unsigned int
        elif ftype == 117:
            val, = unpack('>H', self.input.read(2))
        # 'I': long int
        elif ftype == 73:
            val, = unpack('>i', self.input.read(4))
        # 'i': long unsigned int
        elif ftype == 105:  # 'l'
            val, = unpack('>I', self.input.read(4))
        # 'L': long long int
        elif ftype == 76:
            val, = unpack('>q', self.input.read(8))
        # 'l': long long unsigned int
        elif ftype == 108:
            val, = unpack('>Q', self.input.read(8))
        # 'f': float
        elif ftype == 102:
            val, = unpack('>f', self.input.read(4))
        # 'd': double
        elif ftype == 100:
            val = self.read_float()
        # 'D': decimal
        elif ftype == 68:
            d = self.read_octet()
            n, = unpack('>i', self.input.read(4))
            val = Decimal(n) / Decimal(10 ** d)
        # 'F': table
        elif ftype == 70:
            val = self.read_table()  # recurse
        # 'A': array
        elif ftype == 65:
            val = self.read_array()
        # 't' (bool)
        elif ftype == 116:
            val = self.read_bit()
        # 'T': timestamp
        elif ftype == 84:
            val = self.read_timestamp()
        # 'V': void
        elif ftype == 86:
            val = None
        else:
            raise FrameSyntaxError('Unknown value in table: {!r} ({!r})'.format(ftype, type(ftype)))
        return val

    def read_array(self):
        array_length = unpack('>I', self.input.read(4))[0]
        array_data = AMQPReader(self.input.read(array_length))
        result = []
        while array_data.input.tell() < array_length:
            val = array_data.read_item()
            result.append(val)
        return result

    def read_timestamp(self):
        """Read and AMQP timestamp, which is a 64-bit integer representing seconds since the Unix epoch in 1-second
        resolution

        Return as a Python datetime.datetime object, expressed as localtime.
        """
        return datetime.fromtimestamp(self.read_longlong())


class AMQPWriter:
    """Convert higher-level AMQP types to bytestreams
    """

    def __init__(self, dest=None):
        """
        Note: dest must also implement `getvalue()`, such as :class:`io.BytesIO`

        :param dest: io.BytesIO object, or None create one
        :type dest: io.IOBase or None
        """
        if isinstance(dest, io.BytesIO):
            self.out = dest
        elif dest is None:
            self.out = io.BytesIO()
        else:
            raise TypeError('AMQPWriter needs an `io.BytesIO` or `None`')

        self.bits = []
        self.bitcount = 0

    def _flushbits(self):
        if self.bits:
            out = self.out
            for b in self.bits:
                out.write(pack('B', b))
            self.bits = []
            self.bitcount = 0

    def close(self):
        """Pass through if possible to any file-like destinations
        """
        self.out.close()

    def flush(self):
        """Pass through if possible to any file-like destinations
        """
        self.out.flush()

    def getvalue(self):
        """Get what's been encoded so far if we're working with a BytesIO

        :return: bytes
        :rtype: bytes
        """
        self._flushbits()
        return self.out.getvalue()

    def write(self, b):
        """Write bytes

        :param b: bytes to write
        :type b: bytes
        """
        self._flushbits()
        self.out.write(b)

    def write_bit(self, b):
        """Write a boolean value
        """
        b = 1 if b else 0
        shift = self.bitcount % 8
        if shift == 0:
            self.bits.append(0)
        self.bits[-1] |= (b << shift)
        self.bitcount += 1

    def write_octet(self, n):
        """Write an integer as an unsigned 8-bit value
        """
        if n < 0 or n > 255:
            raise FrameSyntaxError(
                'Octet {0!r} out of range 0..255'.format(n))
        self._flushbits()
        self.out.write(pack('B', n))

    def write_short(self, n):
        """Write an integer as an unsigned 16-bit value
        """
        if n < 0 or n > 65535:
            raise FrameSyntaxError(
                'Octet {0!r} out of range 0..65535'.format(n))
        self._flushbits()
        self.out.write(pack('>H', int(n)))

    def write_long(self, n):
        """Write an integer as an unsigned2 32-bit value
        """
        if n < 0 or n >= 4294967296:
            raise FrameSyntaxError(
                'Octet {0!r} out of range 0..2**31-1'.format(n))
        self._flushbits()
        self.out.write(pack('>I', n))

    def write_longlong(self, n):
        """Write an integer as an unsigned 64-bit value
        """
        if n < 0 or n >= 18446744073709551616:
            raise FrameSyntaxError(
                'Octet {0!r} out of range 0..2**64-1'.format(n))
        self._flushbits()
        self.out.write(pack('>Q', n))

    def write_shortstr(self, s):
        """Write a string up to 255 bytes long (after any encoding)

        If passed a unicode string, encode with UTF-8.
        """
        self._flushbits()
        if isinstance(s, str):
            s = s.encode('utf-8')
        if len(s) > 255:
            raise FrameSyntaxError(
                'Shortstring overflow ({0} > 255)'.format(len(s)))
        self.write_octet(len(s))
        self.out.write(s)

    def write_longstr(self, s):
        """Write a string up to 2**32 bytes long after encoding

        If passed a unicode string, encode as UTF-8.
        """
        self._flushbits()
        if isinstance(s, str):
            s = s.encode('utf-8')
        self.write_long(len(s))
        self.out.write(s)

    def write_table(self, d):
        """Write out a Python dictionary made of up string keys, and values that are strings, signed integers, Decimal,
        datetime.datetime, or sub-dictionaries following the same constraints
        """
        self._flushbits()
        table_data = AMQPWriter()
        for k, v in d.items():
            table_data.write_shortstr(k)
            table_data.write_item(v, k)
        table_data = table_data.getvalue()
        self.write_long(len(table_data))
        self.out.write(table_data)

    def write_item(self, v, k=None):
        if isinstance(v, (str, bytes)):
            if isinstance(v, str):
                v = v.encode('utf-8')
            self.write(b'S')
            self.write_longstr(v)
        elif isinstance(v, bool):
            self.write(pack('>cB', b't', int(v)))
        elif isinstance(v, float):
            self.write(pack('>cd', b'd', v))
        elif isinstance(v, int):
            self.write(pack('>ci', b'I', v))
        elif isinstance(v, Decimal):
            self.write(b'D')
            sign, digits, exponent = v.as_tuple()
            v = 0
            for d in digits:
                v = (v * 10) + d
            if sign:
                v = -v
            self.write_octet(-exponent)
            self.write(pack('>i', v))
        elif isinstance(v, datetime):
            self.write(b'T')
            self.write_timestamp(v)
            # # FIXME: timezone ?
        elif isinstance(v, dict):
            self.write(b'F')
            self.write_table(v)
        elif isinstance(v, (list, tuple)):
            self.write(b'A')
            self.write_array(v)
        elif v is None:
            self.write(b'V')
        else:
            if k:
                err = 'Table type {!r} for key {!r} not handled by amqpy. [value: {!r}]'.format(type(v), k, v)
            else:
                err = 'Table type {!r} not handled by amqpy. [value: {!r}]'.format(type(v), v)
            raise FrameSyntaxError(err)

    def write_array(self, a):
        array_data = AMQPWriter()
        for v in a:
            array_data.write_item(v)
        array_data = array_data.getvalue()
        self.write_long(len(array_data))
        self.out.write(array_data)

    def write_timestamp(self, v):
        """Write out a Python datetime.datetime object as a 64-bit integer representing seconds since the Unix epoch
        """
        self.out.write(pack('>q', int(mktime(v.timetuple()))))



import uuid
import logging
import sys

import pytest

from .. import Channel, Message, FrameSyntaxError
from ..exceptions import AMQPError, ChannelError, PreconditionFailed, NotFound, AccessRefused
from .conftest import get_server_props

logging.basicConfig(level=logging.DEBUG, stream=sys.stdout, style='{', format='{asctime} {levelname:8} {message}',
                    datefmt='%Y/%m/%d %H:%M:%S')
log = logging.getLogger('amqpy')


class TestExchange:
    def test_exchange_declare_and_delete(self, ch):
        exch_name = 'test_exchange_{}'.format(uuid.uuid4())
        ch.exchange_declare(exch_name, 'direct')
        ch.exchange_delete(exch_name)

    def test_exchange_declare_passive_raises(self, ch):
        with pytest.raises(NotFound):
            exch_name = 'test_exchange_{}'.format(uuid.uuid4())
            ch.exchange_declare(exch_name, 'direct', passive=True)

    def test_exchange_redeclare_different_type_raises(self, ch: Channel, rand_exch):
        """Redeclaring an exchange with a different type should raise
        """
        ch.exchange_declare(rand_exch, 'direct')
        with pytest.raises(AMQPError):
            ch.exchange_declare(rand_exch, 'fanout')

    def test_exchange_redeclare_auto_delete_raises(self, ch: Channel, rand_exch):
        """Redeclaring an exchange with a different `auto_delete` should raise
        """
        ch.exchange_declare(rand_exch, 'direct', auto_delete=False)
        with pytest.raises(AMQPError):
            ch.exchange_declare(rand_exch, 'direct', auto_delete=True)

    def test_exchange_delete_nonexistent_raises(self, ch, rand_exch):
        """Test to ensure that deleting a nonexistent exchange raises `NotFound`

        Note: starting with RabbitMQ 3.2 (?), exchange.delete is an idempotent assertion that the exchange must not
        exist.

             We have made queue.delete into an idempotent assertion that the queue must not exist, in the same way that
             queue.declare asserts that it must. See https://www.rabbitmq.com/specification.html

        This means that the RabbitMQ server will not raise a 404 NOT FOUND channel exception when attempting to
        delete a nonexistent exchange.
        """
        server_props = get_server_props(ch.connection)

        if server_props[0] == 'RabbitMQ' and server_props[1] >= (3, 2, 0):
            ch.exchange_delete(rand_exch)
        else:
            with pytest.raises(NotFound):
                ch.exchange_delete(rand_exch)

    def test_exchange_delete_default(self, ch):
        with pytest.raises(AccessRefused):
            ch.exchange_delete('')

    def test_exchange_delete_in_use(self, ch, rand_exch, rand_queue, rand_rk):
        ch.exchange_declare(rand_exch, 'direct')
        ch.queue_declare(rand_queue)
        ch.queue_bind(rand_queue, rand_exch, rand_rk)

        with pytest.raises(PreconditionFailed):
            ch.exchange_delete(rand_exch, if_unused=True)

    def test_exchange_bind(self, ch):
        """Test exchange binding

        Network configuration is as follows (-> is forwards to : source_exchange -> dest_exchange -> queue The test
        checks that once the message is publish to the destination exchange(funtest.topic_dest) it is delivered to the
        queue.
        """
        test_routing_key = 'unit_test__key'
        dest_exchange = 'funtest.topic_dest_bind'
        source_exchange = 'funtest.topic_source_bind'

        ch.exchange_declare(dest_exchange, 'topic', auto_delete=True)
        ch.exchange_declare(source_exchange, 'topic', auto_delete=True)

        qname, _, _ = ch.queue_declare()
        ch.exchange_bind(dest_exch=dest_exchange, source_exch=source_exchange, routing_key=test_routing_key)
        ch.queue_bind(qname, dest_exchange, routing_key=test_routing_key)

        msg = Message('funtest message', content_type='text/plain', application_headers={'foo': 7, 'bar': 'baz'})

        ch.basic_publish(msg, source_exchange, routing_key=test_routing_key)

        msg2 = ch.basic_get(qname, no_ack=True)
        assert msg == msg2

    def test_exchange_unbind(self, ch):
        dest_exchange = 'funtest.topic_dest_unbind'
        source_exchange = 'funtest.topic_source_unbind'
        test_routing_key = 'unit_test__key'

        ch.exchange_declare(dest_exchange, 'topic', auto_delete=True)
        ch.exchange_declare(source_exchange, 'topic', auto_delete=True)

        ch.exchange_bind(dest_exch=dest_exchange, source_exch=source_exchange, routing_key=test_routing_key)

        ch.exchange_unbind(dest_exch=dest_exchange, source_exch=source_exchange, routing_key=test_routing_key)


class TestQueue:
    def test_queue_delete_nonexistent(self, ch):
        """Test to ensure that deleting a nonexistent queue raises `NotFound`

        Note: starting with RabbitMQ 3.2 (?), queue.delete is an idempotent assertion that the queue must not exist.

            We have made exchange.delete into an idempotent assertion that the exchange must not exist, in the same way
            that exchange.declare asserts that it must. See https://www.rabbitmq.com/specification.html

        This means that the RabbitMQ server will not raise a 404 NOT FOUND channel exception when attempting to
        delete a nonexistent queue.
        """
        server_props = get_server_props(ch.connection)
        if server_props[0] == 'RabbitMQ' and server_props[1] >= (3, 2, 0):
            assert ch.queue_delete('bogus_queue_that_does_not_exist') == 0
        else:
            with pytest.raises(NotFound):
                ch.queue_delete('bogus_queue_that_does_not_exist')

    def test_queue(self, ch):
        my_routing_key = 'funtest.test_queue'
        msg = Message('funtest message', content_type='text/plain', application_headers={'foo': 7, 'bar': 'baz'})

        qname, _, _ = ch.queue_declare()
        ch.queue_bind(qname, 'amq.direct', routing_key=my_routing_key)

        ch.basic_publish(msg, 'amq.direct', routing_key=my_routing_key)

        msg2 = ch.basic_get(qname, no_ack=True)
        assert msg == msg2

    def test_unbind(self, ch):
        my_routing_key = 'funtest.test_queue'

        qname, _, _ = ch.queue_declare()
        ch.queue_bind(qname, 'amq.direct', routing_key=my_routing_key)
        ch.queue_unbind(qname, 'amq.direct', routing_key=my_routing_key)

    def test_unbind_nonexistent(self, ch, rand_exch, rand_queue, rand_rk):
        """Test to ensure that unbinding a nonexistent binding raises `NotFound`

        Note: starting with RabbitMQ 3.2 (?), queue.delete is an idempotent assertion that the queue must not exist.

            We have made exchange.delete into an idempotent assertion that the exchange must not exist, in the same way
            that exchange.declare asserts that it must. See https://www.rabbitmq.com/specification.html

        This means that the RabbitMQ server will not raise a 404 NOT FOUND channel exception when attempting to
        delete a nonexistent queue.
        """
        server_props = get_server_props(ch.connection)

        if server_props[0] == 'RabbitMQ' and server_props[1] >= (3, 2, 0):
            ch.queue_unbind(rand_queue, rand_exch, rand_rk)
        else:
            with pytest.raises(NotFound):
                ch.queue_unbind(rand_queue, rand_exch, rand_rk)


class TestPublish:
    def test_publish(self, ch):
        ch.exchange_declare('funtest.fanout', 'fanout', auto_delete=True)
        msg = Message('funtest message', content_type='text/plain', application_headers={'foo': 7, 'bar': 'baz'})
        ch.basic_publish(msg, 'funtest.fanout')

    def test_publish_large(self, ch):
        """Test sending some extra large messages
        """
        qname, _, _ = ch.queue_declare()

        for multiplier in [100, 1000, 10000, 50000, 100000, 500000]:
            msg = Message('this is a test message' * multiplier, content_type='text/plain',
                          application_headers={'foo': 7, 'bar': 'baz'})

            ch.basic_publish(msg, routing_key=qname)

            msg2 = ch.basic_get(no_ack=True)
            assert msg == msg2

    def test_publish_confirm(self, ch: Channel, rand_exch):
        queue_name = 'test.queue.publish'
        rk = queue_name
        ch.exchange_declare(rand_exch, 'direct', auto_delete=True)
        ch.queue_declare(queue_name)
        ch.queue_bind(queue_name, rand_exch, rk)
        msg = Message('funtest message', content_type='text/plain', application_headers={'foo': 7, 'bar': 'baz'})
        ch.confirm_select()
        ch.basic_publish_confirm(msg, rand_exch, rk)


class TestChannel:
    def test_defaults(self, ch):
        """Test how a queue defaults to being bound to an AMQP default exchange, and how publishing defaults to the
        default exchange, and basic_get defaults to getting from the most recently declared queue, and queue_delete
        defaults to deleting the most recently declared queue
        """
        msg = Message('funtest message', content_type='text/plain', application_headers={'foo': 7, 'bar': 'baz'})

        qname, _, _ = ch.queue_declare()
        ch.basic_publish(msg, routing_key=qname)

        msg2 = ch.basic_get(no_ack=True)
        assert msg == msg2

        n = ch.queue_purge()
        assert n == 0

        n = ch.queue_delete()
        assert n == 0

    def test_encoding(self, ch):
        my_routing_key = 'funtest.test_queue'

        qname, _, _ = ch.queue_declare()
        ch.queue_bind(qname, 'amq.direct', routing_key=my_routing_key)

        # no encoding, body passed through unchanged
        msg = Message('hello world')
        ch.basic_publish(msg, 'amq.direct', routing_key=my_routing_key)
        msg2 = ch.basic_get(qname, no_ack=True)
        assert isinstance(msg2.body, str)
        assert msg2.body == 'hello world'

        # default UTF-8 encoding of unicode body, returned as unicode
        msg = Message('hello world')
        ch.basic_publish(msg, 'amq.direct', routing_key=my_routing_key)
        msg2 = ch.basic_get(qname, no_ack=True)
        assert msg2.properties['content_encoding'] == 'UTF-8'
        assert msg2.body == 'hello world'

        # Explicit latin_1 encoding, still comes back as unicode
        msg = Message('hello world', content_encoding='latin_1')
        ch.basic_publish(msg, 'amq.direct', routing_key=my_routing_key)
        msg2 = ch.basic_get(qname, no_ack=True)
        assert msg2.properties['content_encoding'] == 'latin_1'
        assert msg2.body == 'hello world'

        # plain string with specified encoding comes back as unicode
        msg = Message('hello w\xf6rld', content_encoding='latin_1')
        ch.basic_publish(msg, 'amq.direct', routing_key=my_routing_key)
        msg2 = ch.basic_get(qname, no_ack=True)

        assert msg2.properties['content_encoding'] == 'latin_1'
        assert msg2.body == 'hello w\u00f6rld'

        # plain string (bytes in Python 3.x) with bogus encoding
        # don't really care about latin_1, just want bytes
        test_bytes = 'hello w\xd6rld'.encode('latin_1')
        msg = Message(test_bytes, content_encoding='I made this up')
        ch.basic_publish(msg, 'amq.direct', routing_key=my_routing_key)
        msg2 = ch.basic_get(qname, no_ack=True)
        assert msg2.properties['content_encoding'] == 'I made this up'
        assert isinstance(msg2.body, bytes)
        assert msg2.body == test_bytes

        # turn off auto_decode for remaining tests
        ch.auto_decode = False

        # unicode body comes back as utf-8 encoded str
        msg = Message('hello w\u00f6rld')
        ch.basic_publish(msg, 'amq.direct', routing_key=my_routing_key)
        msg2 = ch.basic_get(qname, no_ack=True)
        assert msg2.properties['content_encoding'] == 'UTF-8'
        assert isinstance(msg2.body, bytes)
        assert msg2.body == 'hello w\xc3\xb6rld'.encode('latin_1')

        # plain string with specified encoding stays plain string
        msg = Message('hello w\xf6rld', content_encoding='latin_1')
        ch.basic_publish(msg, 'amq.direct', routing_key=my_routing_key)
        msg2 = ch.basic_get(qname, no_ack=True)
        assert msg2.properties['content_encoding'] == 'latin_1'
        assert isinstance(msg2.body, bytes)
        assert msg2.body == 'hello w\xf6rld'.encode('latin_1')

        # explicit latin_1 encoding, comes back as str
        msg = Message('hello w\u00f6rld', content_encoding='latin_1')
        ch.basic_publish(msg, 'amq.direct', routing_key=my_routing_key)
        msg2 = ch.basic_get(qname, no_ack=True)
        assert msg2.properties['content_encoding'] == 'latin_1'
        assert isinstance(msg2.body, bytes)
        assert msg2.body == 'hello w\xf6rld'.encode('latin_1')

    def test_survives_channel_error(self, ch):
        with pytest.raises(ChannelError):
            ch.queue_declare('krjqheewq_bogus', passive=True)
        ch.queue_declare('funtest_survive')
        ch.queue_declare('funtest_survive', passive=True)
        assert ch.queue_delete('funtest_survive') == 0

    def test_invalid_header(self, ch):
        """Test sending a message with an unserializable object in the header

        http://code.google.com/p/py-amqplib/issues/detail?id=17
        """
        qname, _, _ = ch.queue_declare()

        msg = Message(application_headers={'test': object()})

        with pytest.raises(FrameSyntaxError):
            ch.basic_publish(msg, routing_key=qname)

    def test_basic_return(self, ch):
        ch.exchange_declare('funtest.fanout', 'fanout', auto_delete=True)

        msg = Message('funtest message', content_type='text/plain', application_headers={'foo': 7, 'bar': 'baz'})

        ch.basic_publish(msg, 'funtest.fanout')
        ch.basic_publish(msg, 'funtest.fanout', mandatory=True)
        ch.basic_publish(msg, 'funtest.fanout', mandatory=True)
        ch.basic_publish(msg, 'funtest.fanout', mandatory=True)
        ch.close()

        # 3 of the 4 messages we sent should have been returned
        assert ch.returned_messages.qsize() == 3



import logging

from .. import Message, AbstractConsumer

from ..exceptions import Timeout

log = logging.getLogger('amqpy')


class Consumer(AbstractConsumer):
    def run(self, msg: Message):
        log.info('{.consumer_tag} received message: {} {}'.format(self, msg.properties, msg.body))
        msg.ack()


class TestConsumer:
    def test_basic_consume(self, conn, ch, rand_exch, rand_queue):
        self.consume_count = 0

        def consumer(msg: Message):
            """Consume message
            """
            global consume_count
            log.info('Received message: {}'.format(msg.body))
            msg.ack()
            self.consume_count += 1

        q = rand_queue
        rk = q
        exch = rand_exch

        log.info('Declare exchange')
        ch.exchange_declare(exch, 'direct')

        log.info('Declare queue')
        ch.queue_declare(q)

        log.info('Bind queue')
        ch.queue_bind(q, exch, rk)

        log.info('Enable publisher confirms')
        ch.confirm_select()

        log.info('Publish messages')
        ch.basic_publish_confirm(Message('Hello, world!', content_type='text/plain'), exch, rk)
        ch.basic_publish_confirm(Message('Hello, world!', content_type='text/plain'), exch, rk)

        log.info('Declare consumer')
        ch.basic_consume(q, callback=consumer)

        log.info('Publish messages')
        ch.basic_publish_confirm(Message('Hello, world!', content_type='text/plain'), exch, rk)
        ch.basic_publish_confirm(Message('Hello, world!', content_type='text/plain'), exch, rk)

        log.info('Begin draining events')
        while True:
            try:
                conn.drain_events(0.1)
            except Timeout:
                break

        assert self.consume_count == 4

    def test_consumer(self, conn, ch, rand_exch, rand_queue):
        q = rand_queue
        rk = q
        exch = rand_exch

        log.info('Declare exchange')
        ch.exchange_declare(exch, 'direct')

        log.info('Declare queue')
        ch.queue_declare(q)

        log.info('Bind queue')
        ch.queue_bind(q, exch, rk)

        log.info('Enable publisher confirms')
        ch.confirm_select()

        log.info('Set QOS')
        ch.basic_qos(prefetch_count=1, a_global=True)

        log.info('Publish messages')
        for i in range(10):
            ch.basic_publish_confirm(Message('{}: Hello, world!'.format(i)), exch, rk)

        log.info('Declare consumer')
        c1 = Consumer(ch, q)
        c1.declare()

        log.info('Begin draining events')
        while True:
            try:
                conn.drain_events(0.1)
            except Timeout:
                break

        assert c1.consume_count == 10



import gc

from .. import Channel


class TestConnection:
    def test_create_channel(self, conn):
        ch = conn.channel(1)
        assert isinstance(ch, Channel)
        assert ch.channel_id == 1

        ch2 = conn.channel()
        assert ch2.channel_id != 1

        ch.close()
        ch2.close()

    def test_close(self, conn):
        """Make sure we've broken various references when closing channels and connections, to help with GC
        """
        # create a channel and make sure it's linked as we'd expect
        ch = conn.channel()
        assert 1 in conn.channels
        assert ch.connection == conn
        assert ch.is_open is True

        # close the channel and make sure the references are broken that we expect
        ch.close()
        assert ch.connection is None
        assert 1 not in conn.channels
        assert ch.callbacks == {}
        assert ch.is_open is False

        # close the connection and make sure the references we expect are gone
        conn.close()
        assert conn.connection is None
        assert conn.channels is None

    def test_gc_closed(self, conn):
        """Make sure we've broken various references when closing channels and connections, to help with GC
        """
        unreachable_before = len(gc.garbage)
        # create a channel and make sure it's linked as we'd expect
        conn.channel()
        assert 1 in conn.channels

        # close the connection and make sure the references we expect are gone.
        conn.close()

        gc.collect()
        gc.collect()
        gc.collect()
        assert unreachable_before == len(gc.garbage)

    def test_gc_forget(self, conn):
        """Make sure the connection gets gc'ed when there is no more references to it
        """
        unreachable_before = len(gc.garbage)

        ch = conn.channel()
        assert 1 in conn.channels

        del ch

        gc.collect()
        gc.collect()
        gc.collect()
        assert unreachable_before == len(gc.garbage)



import uuid
import logging
import sys

import pytest

from .. import Connection


class ColouredFormatter(logging.Formatter):
    RESET = '\x1B[0m'
    RED = '\x1B[31m'
    YELLOW = '\x1B[33m'
    BRGREEN = '\x1B[01;32m'  # grey in solarized for terminals

    def format(self, record, colour=False):
        message = super().format(record)

        if not colour:
            return message

        level_no = record.levelno
        if level_no >= logging.CRITICAL:
            colour = self.RED
        elif level_no >= logging.ERROR:
            colour = self.RED
        elif level_no >= logging.WARNING:
            colour = self.YELLOW
        elif level_no >= logging.INFO:
            colour = self.RESET
        elif level_no >= logging.DEBUG:
            colour = self.BRGREEN
        else:
            colour = self.RESET

        message = colour + message + self.RESET

        return message


class ColouredHandler(logging.StreamHandler):
    def __init__(self, stream=sys.stdout):
        super().__init__(stream)

    def format(self, record, colour=False):
        if not isinstance(self.formatter, ColouredFormatter):
            self.formatter = ColouredFormatter()

        return self.formatter.format(record, colour)

    def emit(self, record):
        stream = self.stream
        try:
            #msg = self.format(record, stream.isatty())
            msg = self.format(record, True)  # force coloured output for pytest

            stream.write(msg)
            stream.write(self.terminator)
            self.flush()
        except Exception:
            self.handleError(record)


h = ColouredHandler()
h.formatter = ColouredFormatter('{asctime} {levelname:8} {message}', '%Y-%m-%d %H:%M:%S', '{')
if sys.version_info >= (3, 3):
    logging.basicConfig(level=logging.DEBUG, handlers=[h])
else:
    # Python 3.2 doesn't support the `handlers` parameter for `logging.basicConfig()`
    logging.basicConfig(level=logging.DEBUG)
    logging.root.handlers[0] = h


def get_server_props(cxn):
    """Parse server properties from `cxn`

    An example result::

        ('RabbitMQ', (3, 3, 5))

    :param cxn: connection object
    :type cxn: amqpy.Connection
    :return: tuple(server product name, version)
    :rtype: tuple(str, tuple(int...))
    """
    product = cxn.server_properties['product']
    version = cxn.server_properties['version'].split('.')
    version = tuple([int(i) for i in version])
    x = cxn.server_capabilities
    assert x

    return product, version


@pytest.fixture(scope='function')
def conn(request):
    """Create a connection

    :return: Connection object
    :rtype: amqpy.Connection
    """
    c = Connection(heartbeat=10)

    def fin():
        c.close()

    request.addfinalizer(fin)
    return c


@pytest.fixture(scope='function')
def ch(request, conn):
    """Create a channel

    :return: Channel object
    :rtype: amqpy.Channel
    """
    chan = conn.channel()

    def fin():
        chan.close()

    request.addfinalizer(fin)
    return chan


@pytest.fixture(scope='function')
def rand_queue():
    """Generate a random queue name

    :return: random queue name
    :rtype: str
    """
    queue_name = 'amqpy.test.queue.{}'.format(uuid.uuid4())
    return queue_name


@pytest.fixture(scope='function')
def rand_exch():
    """Generate a random exchange name

    :return: random exchange name
    :rtype: str
    """
    exch_name = 'amqpy.test.exchange.{}'.format(uuid.uuid4())
    return exch_name


@pytest.fixture(scope='function')
def rand_rk():
    """Generate a random routing key

    :return: random routing key
    :rtype: str
    """
    routing_key = 'amqpy.test.rk.{}'.format(uuid.uuid4())
    return routing_key



import logging
import time
from functools import wraps

from . import compat

compat.patch()  # monkey-patch time.perf_counter

log = logging.getLogger('amqpy')


def synchronized(lock_name):
    """Decorator for automatically acquiring and releasing lock for method call

    This decorator accesses the `lock_name` :class:`threading.Lock` attribute of the instance that the wrapped method is
    bound to. The lock is acquired (blocks indefinitely) before the method is called. After the method has executed,
    the lock is released.

    Decorated methods should not be long-running operations, since the lock is held for the duration of the method's
    execution.

    :param lock_name: name of :class:`threading.Lock` object
    """

    def decorator(f):
        @wraps(f)
        def wrapper(self, *args, **kwargs):
            lock = getattr(self, lock_name)
            acquired = lock.acquire(False)
            if not acquired:
                # log.debug('> Wait to acquire lock for [{}]'.format(f.__qualname__))
                start_time = time.perf_counter()
                lock.acquire()
                tot_time = time.perf_counter() - start_time
                if tot_time > 10:
                    # only log if waited for more than 10s to acquire lock
                    log.debug('> Acquired lock for [{}] in: {:.3f}s'.format(f.__qualname__, tot_time))
            try:
                retval = f(self, *args, **kwargs)
            finally:
                lock.release()
            return retval

        return wrapper

    return decorator



from collections import namedtuple
import struct

from .serialization import AMQPReader, AMQPWriter

queue_declare_ok_t = namedtuple('queue_declare_ok_t', ('queue', 'message_count', 'consumer_count'))

basic_return_t = namedtuple('basic_return_t', ('reply_code', 'reply_text', 'exchange', 'routing_key', 'message'))

method_t = namedtuple('method_t', ('class_id', 'method_id'))

#: The default, minimum frame size that both the client and server must be able to handle
FRAME_MIN_SIZE = 4096


class FrameType:
    """This class contains frame-related constants

    METHOD, HEADER, BODY, and HEARTBEAT are all frame type constants which make up the first byte of every frame. The
    END constant is the termination value which is the last byte of every frame.
    """
    METHOD = 1  # method frame
    HEADER = 2  # content header frame
    BODY = 3  # content body frame
    HEARTBEAT = 8  # heartbeat frame
    END = 206  # not actually a frame type; this is the frame terminator byte


class Connection:
    CLASS_ID = 10

    Start = method_t(10, 10)
    StartOk = method_t(10, 11)
    Secure = method_t(10, 20)
    SecureOk = method_t(10, 21)
    Tune = method_t(10, 30)
    TuneOk = method_t(10, 31)
    Open = method_t(10, 40)
    OpenOk = method_t(10, 41)
    Close = method_t(10, 50)
    CloseOk = method_t(10, 51)
    Blocked = method_t(10, 60)
    Unblocked = method_t(10, 61)


class Channel:
    CLASS_ID = 20

    Open = method_t(20, 10)
    OpenOk = method_t(20, 11)
    Flow = method_t(20, 20)
    FlowOk = method_t(20, 21)
    Close = method_t(20, 40)
    CloseOk = method_t(20, 41)


class Exchange:
    CLASS_ID = 40

    Declare = method_t(40, 10)
    DeclareOk = method_t(40, 11)
    Delete = method_t(40, 20)
    DeleteOk = method_t(40, 21)
    Bind = method_t(40, 30)
    BindOk = method_t(40, 31)
    Unbind = method_t(40, 40)
    UnbindOk = method_t(40, 51)


class Queue:
    CLASS_ID = 50

    Declare = method_t(50, 10)
    DeclareOk = method_t(50, 11)
    Bind = method_t(50, 20)
    BindOk = method_t(50, 21)
    Purge = method_t(50, 30)
    PurgeOk = method_t(50, 31)
    Delete = method_t(50, 40)
    DeleteOk = method_t(50, 41)
    Unbind = method_t(50, 50)
    UnbindOk = method_t(50, 51)


class Basic:
    CLASS_ID = 60

    Qos = method_t(60, 10)
    QosOk = method_t(60, 11)
    Consume = method_t(60, 20)
    ConsumeOk = method_t(60, 21)
    Cancel = method_t(60, 30)
    CancelOk = method_t(60, 31)
    Publish = method_t(60, 40)
    Return = method_t(60, 50)
    Deliver = method_t(60, 60)
    Get = method_t(60, 70)
    GetOk = method_t(60, 71)
    GetEmpty = method_t(60, 80)
    Ack = method_t(60, 80)
    Reject = method_t(60, 90)
    RecoverAsync = method_t(60, 100)
    Recover = method_t(60, 110)
    RecoverOk = method_t(60, 111)


class Confirm:
    CLASS_ID = 85

    Select = method_t(85, 10)
    SelectOk = method_t(85, 11)


class Tx:
    CLASS_ID = 90

    Select = method_t(90, 10)
    SelectOk = method_t(90, 11)
    Commit = method_t(90, 20)
    CommitOk = method_t(90, 21)
    Rollback = method_t(90, 30)
    RollbackOk = method_t(90, 31)


class Frame:
    """AMQP frame

    The format of the AMQP frame is as follows::

        offset:         0      1         3         7                 size+7      size+8
                        +------+---------+---------+-------------------+-----------+
                        | type | channel |  size   |  --- payload ---  | frame-end |
                        +------+---------+---------+-------------------+-----------+
        size (bytes)        1       2         4             size             1
    """

    def __init__(self, frame_type=None, channel=0, payload=bytes()):
        """Create new Frame

        Leave all three parameters as default to create an empty frame whose `data` can be manually written to
        afterwards.

        :param frame_type: frame type
        :param channel: associated channel number
        :param payload: frame payload
        :type frame_type: int
        :type channel: int
        :type payload: bytes or bytearray
        """
        #: raw frame data; can be manually manipulated at any time
        self.data = bytearray()

        self._frame_type = None
        self._channel = None
        self._payload_size = None

        # create bytearray from provided data
        if frame_type is not None:
            self._frame_type = frame_type
            self._channel = channel
            self._payload_size = len(payload)
            frame_format = '>BHI{}sB'.format(self._payload_size)
            self.data = struct.pack(frame_format, frame_type, channel, self._payload_size, payload, FrameType.END)

    @property
    def frame_type(self):
        """Get frame type

        :return: frame type
        :rtype: int
        """
        if self._frame_type is not None:
            return self._frame_type
        else:
            self._frame_type = struct.unpack_from('>B', self.data)[0]
            return self._frame_type

    @property
    def channel(self):
        """Get frame channel number

        :return: channel number
        :rtype: int
        """
        if self._channel is not None:
            return self._channel
        else:
            self._channel = struct.unpack_from('>H', self.data, 1)[0]
            return self._channel

    @property
    def payload_size(self):
        """Get frame payload size

        :return: payload size
        :rtype: int
        """
        if self._payload_size is not None:
            return self._payload_size
        else:
            self._payload_size = struct.unpack_from('>I', self.data, 3)[0]
            return self._payload_size

    @property
    def payload(self):
        """Get frame payload

        :return: payload
        :rtype: bytearray
        """
        return self.data[7:-1]


class Method:
    """AMQP method

    The format of the method frame `FrameType.METHOD` payload is as follows::

        offset:         0          2           4
                        +----------+-----------+-------------- - -
                        | class-id | method-id | arguments...
                        +----------+-----------+-------------- - -
        size (bytes):         2          2           ...
    """

    def __init__(self, method_type, args=None, content=None, channel_id=None):
        """
        :param method_type: method type
        :param args: method args
        :param content: content
        :param channel_id: the associated channel ID, if any
        :type method_type: method_t
        :type args: AMQPReader or AMQPWriter or None
        :type content: amqp.message.GenericContent or None
        :type channel_id: int or None
        """
        self.method_type = method_type
        assert isinstance(method_type, method_t)
        if isinstance(args, (AMQPReader, AMQPWriter)):
            self.args = args
        elif args is None:
            self.args = AMQPWriter()
        else:
            raise ValueError('args must be an instance of `AMQPReader` or `AMQPWriter`')
        self.content = content
        self.channel_id = channel_id

        self._body_bytes = bytes()  # used internally to store encoded body

    def pack_method(self):
        """Pack this method into a bytes object suitable for using as a payload for `FrameType.METHOD` frames

        :return: bytes
        :rtype: bytes
        """
        return struct.pack('>HH', self.method_type.class_id, self.method_type.method_id) + self.args.getvalue()

    def pack_header(self):
        """Pack this method into a bytes object suitable for using as a payload for `FrameType.HEADER` frames

        :return: bytes
        :rtype: bytes
        """
        if not self.content:
            raise ValueError('`pack_header()` is only meaningful if there is content to pack')

        self._body_bytes = self.content.body
        if isinstance(self._body_bytes, str):
            # encode body to bytes
            coding = self.content.properties.setdefault('content_encoding', 'UTF-8')
            self._body_bytes = self.content.body.encode(coding)
        properties = self.content.serialize_properties()
        return struct.pack('>HHQ', self.method_type.class_id, 0, len(self._body_bytes)) + properties

    def pack_body(self, chunk_size):
        """Pack this method into a bytes object suitable for using as a payload for `FrameType.BODY` frames

        :param chunk_size: split up body into pices that are at most `chunk_size` bytes each
        :type chunk_size: int
        :return: bytes generator
        :rtype: generator[bytes]
        """
        if not self.content:
            raise ValueError('`pack_body()` is only meaningful if there is content to pack')

        for i in range(0, len(self._body_bytes), chunk_size):
            yield self._body_bytes[i:i + chunk_size]



"""Code common to Connection and Channel objects
"""
from abc import ABCMeta, abstractmethod
from threading import Lock

from .exceptions import AMQPNotImplementedError, RecoverableConnectionError
from .spec import Method
from . import spec

__all__ = ['AbstractChannel']


class AbstractChannel(metaclass=ABCMeta):
    """Superclass for both the Connection, which is treated as channel 0, and other user-created Channel objects

    The subclasses must have a METHOD_MAP class variable, mapping between AMQP method signatures and Python methods.
    """

    #: Placeholder, implementations must override this
    METHOD_MAP = {}

    #: List of methods which must be handled immediately
    IMMEDIATE_METHODS = [spec.Basic.Return]

    def __init__(self, connection, channel_id):
        """
        :type connection: amqpy.connection.Connection
        :type channel_id: int
        """
        self.connection = connection
        self.channel_id = channel_id
        connection.channels[channel_id] = self
        # queue of incoming methods for this channel
        self.method_queue = []  # list[Method]
        self.auto_decode = False
        self.lock = Lock()

    @abstractmethod
    def close(self):
        """Close this Channel or Connection
        """
        pass

    def __enter__(self):
        return self

    # noinspection PyUnusedLocal
    def __exit__(self, exc_type, exc_value, traceback):
        self.close()

    def _send_method(self, method):
        if self.connection is None:
            raise RecoverableConnectionError('connection already closed')

        self.connection.method_writer.write_method(self.channel_id, method)

    def wait(self, allowed_methods=None, callback=None):
        """Wait for a method that matches any one of `allowed_methods`

        If `callback` is specified, `callback(channel, method)` will be called after a method has been received. If
        `callback` is `None`, the default callback for the received method will be called (as specified in the channel's
        `METHOD_MAP`).

        :param allowed_methods: list of possible methods to wait for, or `None` to wait for any method
        :param callback: callable with the following signature: callable(AbstractChannel, Method)
        :type allowed_methods: list or None
        :type callback: Callable(AbstractChannel, Method)
        """
        method = self._wait_method(allowed_methods)
        return self.handle_method(method, callback=callback)

    def _wait_method(self, allowed_methods):
        """Wait for a method from the server destined for the current channel

        This method is designed to be called from a channel instance.

        :return: method
        :rtype: Method
        """
        # create a more convenient list of methods to check
        if isinstance(allowed_methods, list):
            # we should always check of the incoming method is `Channel.Close`
            allowed_methods = [spec.Channel.Close] + allowed_methods

        # check the channel's method queue
        method_queue = self.connection.channels[self.channel_id].method_queue

        for qm in method_queue:
            assert isinstance(qm, Method)
            if allowed_methods is None or qm.method_type in allowed_methods:
                # found the method we're looking for in the queue
                method_queue.remove(qm)
                return qm

        # nothing queued, need to wait for a method from the server
        while True:
            method = self.connection.method_reader.read_method()
            ch_id = method.channel_id
            m_type = method.method_type

            # check if the received method is the one we're waiting for
            if method.channel_id == self.channel_id and (allowed_methods is None or m_type in allowed_methods):
                # received the method we're waiting for
                return method

            # check if the received method needs to be handled immediately
            if ch_id != 0 and m_type in self.IMMEDIATE_METHODS:
                # certain methods like basic_return should be dispatched immediately rather than being queued, even if
                # they're not one of the `allowed_methods` we're looking for
                self.connection.channels[ch_id].handle_method(method)
                continue

            # not the channel and/or method we were looking for; queue this method for later
            self.connection.channels[ch_id].method_queue.append(method)

            # if the method is destined for channel 0 (the connection itself), it's probably an exception, so handle it
            # immediately
            if ch_id == 0:
                self.connection.wait()

    def _wait_multiple(self, channels, allowed_methods, timeout=None):
        """Wait for an event on multiple channels

        It only makes sense for the `Connection` instance to call this method (as opposed to `Channel` instances).
        This method gets called by :meth:`Connection.drain_events()`.

        :param channels: dict of channels to watch
        :param allowed_methods: list of allowed methods
        :param timeout: timeout
        :type channels: dict[ch_id int: channel Channel]
        :return: tuple(channel_id, method)
        :rtype: tuple(int, Method)
        """
        # create a more convenient list of methods to check
        if isinstance(allowed_methods, list):
            # we should always check of the incoming method is `Channel.Close`
            allowed_methods = [spec.Channel.Close] + allowed_methods

        # check the method queue of each channel
        for ch_id, channel in channels.items():
            for qm in channel.method_queue:
                assert isinstance(qm, Method)
                if allowed_methods is None or qm.method_type in allowed_methods:
                    channel.method_queue.remove(qm)
                    return ch_id, qm

        # nothing queued for any channel, need to wait for a method from the server
        while True:
            # do a blocking read for any incoming method
            method = self.connection.method_reader.read_method(timeout)
            m_type = method.method_type
            ch_id = method.channel_id

            # check if the received method is the one we're waiting for
            if ch_id in channels and (allowed_methods is None or m_type in allowed_methods):
                return ch_id, method

            # check if the received method needs to be handled immediately
            if ch_id != 0 and m_type in self.IMMEDIATE_METHODS:
                # certain methods like basic_return should be dispatched immediately rather than being queued, even if
                # they're not one of the `allowed_methods` we're looking for
                self.connection.channels[ch_id].handle_method(method)
                continue

            # not the channel and/or method we were looking for; queue this method for later
            channels[ch_id].method_queue.append(method)

            # if the method is destined for channel 0 (the connection itself), it's probably an exception, so handle it
            # immediately
            if ch_id == 0:
                self.connection.wait()

    def handle_method(self, method, channel=None, callback=None):
        """Handle the specified received method

        The appropriate handler as defined in `self.METHOD_MAP` will be called to handle this method.

        :param method: freshly received method from the server
        :param channel: channel object
        :param callback: callable with the following signature: callable(AbstractChannel, Method)
        :type method: amqpy.spec.Method
        :type channel: amqpy.channel.Channel
        :type callback: Callable(AbstractChannel, Method) or None
        :return: the return value of the specific callback or method handler
        """
        #: :type: GenericContent
        content = method.content
        channel = channel or self

        if content and channel.auto_decode and 'content_encoding' in content.properties:
            # try to decode message body
            # noinspection PyBroadException
            try:
                content.body = content.body.decode(content.properties['content_encoding'])
            except Exception:
                pass

        try:
            callback = callback or channel.METHOD_MAP[method.method_type]
        except KeyError:
            raise AMQPNotImplementedError('Unknown AMQP method {0}'.format(method.method_type))

        return callback(channel, method)



from abc import ABCMeta, abstractmethod
from threading import Thread


class AbstractConsumer(metaclass=ABCMeta):
    """
    This class provides facilities to create and manage queue consumers. To create a consumer, subclass this class
    and override the :meth:`run` method. Then, instantiate the class with the desired parameters and call
    :meth:`declare` to declare the consumer to the server.

    Example::

        class Consumer(AbstractConsumer):
            def run(self, msg: Message):
                print('Received message: {}'.format(msg.body))
                msg.ack()

        c1 = Consumer(ch, 'test.q')
        c1.declare()

        conn.drain_events()

    To take full advantage of the work queue and consumer architecture, specify `use_thread=True` and make sure that QOS
    settings are enabled for the channel or connection. This way, multiple consumers may be declared on the same
    channel, and messages will be load-balanced among them.
    """

    def __init__(self, channel, queue, consumer_tag='', no_local=False, no_ack=False, exclusive=False,
                 use_thread=False):
        """
        :param channel: channel
        :type channel: amqpy.channel.Channel
        :param str queue: queue
        :param str consumer_tag: consumer tag, local to the connection; leave blank to let server auto-assign a tag
        :param bool no_local: if True: do not deliver own messages
        :param bool no_ack: server will not expect an ack for each message
        :param bool exclusive: request exclusive access
        :param bool use_thread: when the consumer run() method is called, start it in a new thread
        """
        self.channel = channel
        self.queue = queue
        self.consumer_tag = consumer_tag
        self.no_local = no_local
        self.no_ack = no_ack
        self.exclusive = exclusive
        self.use_thread = use_thread

        #: Number of messages consumed (incremented automatically)
        self.consume_count = 0

    def declare(self):
        """Declare the consumer

        This method calls :meth:`~amqpy.channel.Channel.basic_consume()` internally.

        After the queue consumer is created, :attr:`self.consumer_tag` is set to the server-assigned consumer tag if
        a tag was not specified initially.
        """
        self.consumer_tag = self.channel.basic_consume(self.queue, self.consumer_tag, self.no_local, self.no_ack,
                                                       self.exclusive, callback=self.start, on_cancel=self.cancel_cb)

    def cancel(self):
        """Cancel the consumer
        """
        self.channel.basic_cancel(self.consumer_tag)

    @abstractmethod
    def run(self, msg):
        """Consumer callback

        This method is called when the consumer is delivered a message. This method may be overridden in the subclass.

        :param msg: received message
        :type msg: amqpy.message.Message
        """
        pass

    def cancel_cb(self, consumer_tag):
        """Consumer cancel callback

        This method is called when the consumer is cancelled. This method **must** be overridden in the subclass.

        :param str consumer_tag: consumer tag
        """
        pass

    def _run(self, msg):
        self.run(msg)
        self.consume_count += 1

    def start(self, msg):
        if self.use_thread:
            t = Thread(target=self._run, args=(msg,), name='consumer-{}'.format(self.consumer_tag))
            t.start()
        else:
            self._run(msg)
